
// BarsClock.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th

/* Start a clock thread which displays the current time on the
   board, and keeps updating. When the user presses enter, main()
   will wake up and kill the thread.

   Unlike Clock.java, display the time as three lines, inspired 
   by the "Matrix M6001" LCD watch (https://www.slashgear.com/matrix-m6001-lcd-watch-169699/)
*/


import java.util.*;
import java.time.*;
import java.time.temporal.*;


public class BarsClock extends Thread
{
  private static final int REFRESH_DELAY = 1000*20;  // 20 secs in ms

  private volatile boolean isRunning = false;   // used to stop the thread
  private Screen scr;



  public BarsClock(Screen scr)
  { this.scr = scr;
    this.start();
  }


  public void run()
  {
    isRunning = true;
    while(isRunning) {
      scr.clear();
      addTime(scr);
      scr.draw();
      Utils.delay(REFRESH_DELAY);
    }
  } // end of run()


  public void end()
  {  isRunning = false;   }



  private void addTime(Screen scr)
  /* Get the current time in hours and minutes, and
     convert the minutes value into two integers
     representing the number of 5 mins, and the remainder.

     Draw three lines using the three integers, and
     add some dots to make it easier to visually judge the
     length of long lines.
  */
  {
    Calendar cal = new GregorianCalendar();
    int hours = cal.get(Calendar.HOUR); // 0..11
    int mins = cal.get(Calendar.MINUTE); // 0..59
    int num5s = mins/5;
    int units = mins - (5*num5s);
    // System.out.println(hours + " : " + num5s + "+" + units);

    // draw hours line
    scr.addLine(0,0, hours-1, 0);
    if (hours > 4) 
      scr.set(4,1, true);    // to mark 5 hours
    if (hours > 9)
      scr.set(9,1, true);    // to mark 10 hours

    // draw 5 mins line
    if (num5s > 0)
      scr.addLine(0,3, num5s-1, 3);
    if (num5s > 6)         // to mark 6 lots of 5 mins i.e. 30 mins
      scr.set(5,4, true);

    if (units > 0)     // draw units line
      scr.addLine(0,6, units-1, 6);
  }  // end of addTime()



  // -------------------------------------

  public static void main(String[] args)
  {
    Screen scr = new Screen();
    // sprites store not needed

    BarsClock c = new BarsClock(scr);

    Utils.waitEnter();    // look at the clock for a bit...
    c.end();
    scr.close();
  }  // end of main()


}  // end of BarsClock class

