
// ArrUtils.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th


/* Utilities for manipulating 2D char arrays:

    * create a list of arrays from one array
    * transform a array into a new array
        - copy  flipHoriz  flipVert  invert
          rotLeft  rotRight
    * clearing an array
    * char array containment
    * printing an array
    * extending a char array
    * copying part of an array
    * trimming blank columns from an array's ends
*/

import java.awt.*;
import java.util.*;
import java.lang.reflect.*;  



public class ArrUtils
{


  // ----- create a list of char arrays from one array ----


  public static ArrayList<char [][]> buildUp(char[][] charArr)
  /* Create a list by generating multiple arrays from the 
     supplied character array.
     Each new array is a slice of the input array by increasing 
     the number of rows in the slice counting from the 
     bottom of the array.
  */
  { if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }

    int width = charArr[0].length;
    int height = charArr.length;
    ArrayList<char [][]> pics = new ArrayList<>();
    for (int rowStart = height-1; rowStart >= 0; rowStart--) {
      char[][] out = new char[height][width];
      clear(out);
      for (int r = rowStart; r < height; r++) {
        for (int c = 0; c < width; c++)
          out[r][c] = charArr[r][c];
      }
      pics.add(out);
    }

    // System.out.println("Pics size: " + pics.size());
    return pics;
  }  // end of buildUp()




  public static ArrayList<char [][]> buildRight(char[][] charArr)
  /* Create a list by generating multiple arrays from the 
     supplied character array.
     Each new array is a slice of the input array by increasing 
     the number of columns in the slice counting from the 
     left of the array.
  */
  { if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }

    int width = charArr[0].length;
    int height = charArr.length;
    ArrayList<char [][]> pics = new ArrayList<>();
    for (int colEnd = 0; colEnd < width; colEnd++) {
      char[][] out = new char[height][width];
      clear(out);
      for (int r = 0; r < height; r++) {
        for (int c = 0; c <= colEnd; c++)
          out[r][c] = charArr[r][c];
      }
      pics.add(out);
    }

    // System.out.println("Pics size: " + pics.size());
    return pics;
  }  // end of buildRight()




  // ------- transform a char array into a new array ----------


  public static char[][] transform(String methodNm, char[][] charArr)
  /* The supplied character array is transformed into a new array
     by calling the function named in methodNm. The function must
     be in this class.
     Currently the functions are: copy, flipHoriz, flipVert, invert,
     rotLeft, rotRight.
  */
  {
    char[][] out = charArr;
    try {
      Class<?> c = Class.forName("ArrUtils");
            // only the methods in this class are considered
      Class<?>[] paramTypes = { char[][].class };
      Object[] params = { charArr };

      Method method = c.getDeclaredMethod(methodNm, paramTypes);
      out = (char[][]) method.invoke(null, params);
    }
    catch(Exception e)
    { System.out.println("Unable to apply transformation: " + methodNm);
      System.out.println(e);
    }
    return out;
  }  // end of transform()


  public static ArrayList<String> getTransforms()
  /* Use reflection to get a list of all the transformation method
     names in this class. A transformation function takes a single
     char array as input and returns a new char array.
  */
  {
    ArrayList<String> names = new ArrayList<String>();
    try {
      Class<?> c = Class.forName("ArrUtils");
      Method[] methods = c.getMethods();
      for (Method m : methods) {
        if (m.getParameterCount() == 1) {
          Class[] types = m.getParameterTypes();
          Class<?> retType = m.getReturnType();
          if ((types[0].equals(char[][].class)) &&
              (retType.equals(char[][].class)) )
            names.add(m.getName());
        }
      }
    }
    catch(Exception e)
    { System.out.println("Unable to collect transformation names");
      System.out.println(e);
    }

    Collections.sort(names);
    return names;
  }  // end of getTransforms()


  public static void printTransforms()
  {
    ArrayList<String> nms = getTransforms();
    if (nms.size() == 0)
      System.out.println("No transformation names found");
    else {
      System.out.println("Transformation names:");
      int i=0;
      for(String nm : nms) {
        System.out.print("  " + nm);
        i = (i+1)%4;
        if (i == 0)
          System.out.println();
      }
      System.out.println();
    } 
  }  // end of printTransforms()



  // transformation functions -----------


  public static char[][] copy(char[][] charArr)
  // copy the array without change
  {
    if (charArr == null) {
      // System.out.println("Array is null");
      return null;
    }
    int width = charArr[0].length;
    int height = charArr.length;
    char[][] out = new char[height][width];
    for (int r = 0; r < height; r++)
       for (int c = 0; c < width; c++)
         out[r][c] = charArr[r][c];
    return out;
  }  // end of copyArray()



  public static char[][] rotRight(char[][] charArr) 
  /* create a new char array which rotates the values in
     the input array clockwise by 90 degrees */
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }
    int width = charArr[0].length;
    int height = charArr.length;
    char[][] out = new char[width][height];

    for (int r = 0; r < height; r++)
      for (int c = 0; c < width; c++)
        out[c][height-1-r] = charArr[r][c];

    return out;
  }  // end of rotRight()



  public static char[][] rotLeft(char[][] charArr) 
  /* create a new char array which rotates the values in
     the input array counter-clockwise by 90 degrees */
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }
    int width = charArr[0].length;
    int height = charArr.length;
    char[][] out = new char[width][height];

    for (int r = 0; r < height; r++)
      for (int c = 0; c < width; c++)
        out[width-1-c][r] = charArr[r][c]; 

    return out;
  }  // end of rotLeft()



  public static char[][] flipVert(char[][] charArr) 
  /* create a new char array which vertically flips 
     the values in the input array */
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }

    int width = charArr[0].length;
    int height = charArr.length;

    char[][] out = new char[height][width];
    for (int r = 0; r < height; r++)
      for (int c = 0; c < width; c++)
        out[r][width-1-c] = charArr[r][c];

    return out;
  }  // end of flipVert()



  public static char[][] flipHoriz(char[][] charArr) 
  /* create a new char array which horizontally flips 
     the values in the input array */
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }

    int width = charArr[0].length;
    int height = charArr.length;

    char[][] out = new char[height][width];
    for (int r = 0; r < height; r++)
      for (int c = 0; c < width; c++)
        out[height-1-r][c] = charArr[r][c];

    return out;
  }  // end of flipHoriz()



  public static char[][] invert(char[][] charArr) 
  /* create a new char array which inverts 
     the values in the input array (i.e. on <--> off */
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }

    int width = charArr[0].length;
    int height = charArr.length;

    char[][] out = new char[height][width];
    for (int r = 0; r < height; r++)
      for (int c = 0; c < width; c++)
        out[r][c] = ((charArr[r][c] == ' ') ? '*' : ' ');

    return out;
  }  // end of invert()



  // ------------ clearing an array ---------


  public static void clear(char[][] charArr) 
  // fill the array with spaces
  {
    if (charArr != null) {
      int width = charArr[0].length;
      int height = charArr.length;
      for (int r = 0; r < height; r++)
        for (int c = 0; c < width; c++)
          charArr[r][c] = ' ';
    }
  }  // end of clear()



  public static void clear(char[][] charArr, Rectangle r)
  {  clear(charArr, r.x, r.y, r.width, r.height);  }


  public static void clear(char[][] charArr, int x, int y, int w,  int h)
  // clear the rectangle region (w x h) starting at (x,y)
  {
    if (!inside(charArr,x,y))
      return;

    int width = charArr[0].length;
    int height = charArr.length;
    int xEnd = (x+w > width) ? width : (x+w);
    int yEnd = (y+h > height) ? height : (y+h);

    for (int r = y; r < yEnd; r++)
      for (int c = x; c < xEnd; c++)
        charArr[r][c] = ' ';
  }  // end of clear()



  // ----- char array containment -----------------


  public static boolean inside(char[][] charArr, Point p)
  {  return inside(charArr, p.x, p.y);  }



  public static boolean inside(char[][] charArr, int x, int y)
  {  
    if (charArr == null)
      return false;

    int width = charArr[0].length;
    int height = charArr.length;
    return ((x >= 0) && (x < width) &&
            (y >= 0) && (y < height));
  }  // end of inside()




  // --------- printing an array -------------


  public static void print(char[][] charArr)
  /* print the array with a border around it, and 
     preceded by its size */
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return;
    }
    int width = charArr[0].length;
    int height = charArr.length;
    System.out.println("Size (w x h): " + 
                              width + " x " + height);
    printLine(width+2);
    for (int r = 0; r < height; r++) {
       System.out.print("|");
       for (int c = 0; c < width; c++)
          System.out.print(charArr[r][c]);
       System.out.println("|");
    }  
    printLine(width+2);
    System.out.println();
  }  // end of print()


  private static void printLine(int no)
  { for (int i=0; i < no; i++)
      System.out.print("-");
    System.out.println();
  }


  // ------------- Extending a Character Array -----------



  public static boolean add(char[][] out, int x, int y, char[][] charArr)
  /* add the char array to the out array with its top-left
     corner placed at (x,y)
     Ignoring excessive rows/columns in the char array.
  */
  {
    if (charArr == null) {
      System.out.println("Input array is null");
      return false;
    }
    if (out == null) {
      System.out.println("Output array is null");
      return false;
    }
    if (!inside(out, x, y)) {
      System.out.println("Coordinate is not inside output array");
      return false;
    }

    int caWidth = charArr[0].length;
    int caHeight = charArr.length;
    int width = out[0].length;
    int height = out.length;

    if ((x+caWidth) > width) {
      System.out.println("Input array has too many columns; ignoring some");
      caWidth = width - x;
    }

    if ((y+caHeight) > height) {
      System.out.println("Input array has too many rows; ignoring some");
      caHeight = height - y;
    }

    for (int r = 0; r < caHeight; r++) {
      int xPos = x;
      for (int c = 0; c < caWidth; c++) {
        if (charArr[r][c] == '*')
          out[y][xPos] = '*';    // only draw '*'s
        xPos++;
      }
      y++;
    }
    return true;
  }  // end of add()



  public static boolean append(char[][] out, int x, char[][] charArr)
  /* Add the contents of charArr to the out char array
     starting at its column, x. 
  */
  { return add(out, x, 0, charArr);  }



  // ------ copying part of a char array -------------------


  public static char[][] copyRegion(Rectangle rect, Point pos, 
                                                      char[][] charArr)
  /* Copy the contents of the char array that is covered by the rectangle.
     The rectangle's (x,y) is in screen coordinates, and the top-left
     corner of the array (i.e. charArr[0][0]) is located at pos.
  */
  {
    if ((rect.width <= 0) || (rect.height <= 0)) {
      // System.out.println("Region is empty");
      return null;
    }
    if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }

    int colStart = rect.x - pos.x;
    int rowStart = rect.y - pos.y;
    if ((colStart < 0) || (colStart > charArr[0].length) ||
        (rowStart < 0) || (rowStart > charArr.length)) {
      System.out.println("Rectangle does not overlap the char array's position");
      return null;
    }

    int colEnd = colStart+rect.width;
    if (colEnd > charArr[0].length) {
      System.out.println("Rectangle is too wide");
      colEnd = charArr[0].length;
    }

    int rowEnd = rowStart+rect.height;
    if (rowEnd > charArr.length) {
      System.out.println("Rectangle is too tall");
      rowEnd = charArr.length;
    }

    char[][] out = new char[rowEnd-rowStart][colEnd-colStart];
    int y = 0;
    for (int r = rowStart; r < rowEnd; r++) {
      int x = 0;
      for (int c = colStart; c < colEnd; c++) {
        out[y][x] = charArr[r][c];
        x++;
      }
      y++;
    }
    return out;
  }  // end of copyRegion()



  // ------------- trimming ------------------


  public static char[][] trim(char[][] charArr)
  /* copy the char array, removing the 
     starting and ending blank columns */
  { 
    if (charArr == null) {
      System.out.println("Array is null");
      return null;
    }

    int height = charArr.length;

    int startCol = getFirstTextCol(charArr);
    int endCol = getLastTextCol(charArr);
    int width = endCol+1-startCol;
    // System.out.println("width: " + width);

    if (width <= 0) {
      // copy the first column instead
      startCol = 0;    
      endCol = 0;
      width = 1;
    }

    // copy char array between startCol and endCol into out
    char[][] out = new char[height][width];
    ArrUtils.clear(out);
    for (int row = 0; row < height; row++) {
      int pos = 0;
      for (int c = startCol; c <= endCol; c++) {
        out[row][pos] = charArr[row][c];
        pos++;
      }
    }
    return out;
  }  // end of trim()



  public static int getFirstTextCol(char[][] charArr)
  /* return the index of the first column with text
     starting from the left  */
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return 0;
    }
    int col = 0;
    while (isBlankCol(charArr, col))
      col++;
    return col;
  }  // end of getFirstTextCol()


  public static int getLastTextCol(char[][] charArr)
  /* return the index of the last column with text
     starting from the right  */
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return 0;
    }
    int width = charArr[0].length;
    int col = width-1;
    while (isBlankCol(charArr, col))
      col--;
    return col;
  }  // end of getLastTextCol()



  public static boolean isBlankCol(char[][] charArr, int col)
  // is this column of the char array all spaces?
  {
    if (charArr == null) {
      System.out.println("Array is null");
      return false;
    }
    int width = charArr[0].length;
    int height = charArr.length;
    if ((col < 0) || (col >= width))
      return false;

    for (int r = 0; r < height; r++)
      if (charArr[r][col] != ' ')
        return false;
    return true;
  }  // end of isBlankCol()




}  // end of Utils class