
// DrawPanel.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th

/* A drawing pane that mimics the Dream Cheeky LEF Message Board,
   to make it a bit easier to design sprites.

   The screen is represented by a 2D array of booleans.

   Each LED is drawn as a rectangle, either on or off
*/

import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.awt.geom.*;



public class DrawPanel extends JPanel
{
  // dimensions of the screen in terms of number of LED lights
  private static final int SCR_WIDTH = 21;
  private static final int SCR_HEIGHT = 7;

  // pixel dimensions of a LED light 
  // (double-sized compared the numbers used in LedScreen.java)
  private static final int LIGHT_WIDTH = 20;
  private static final int LIGHT_HEIGHT = 14;

  // distance between each LED
  private static final int X_OFFSET = 5;
  private static final int Y_OFFSET = 5;   // same spacing

  // distance between LEDs and the pane's edges
  private static final int EDGE_OFFSET = 10;

  // distance from one LED to the next
  private static final int X_STEP = LIGHT_WIDTH + X_OFFSET;
  private static final int Y_STEP = LIGHT_HEIGHT + Y_OFFSET;


  // pixel dimensions of the pane
  private static final int BOARD_WIDTH = 
                              EDGE_OFFSET + (SCR_WIDTH * X_STEP);
  private static final int BOARD_HEIGHT = 
                              EDGE_OFFSET + (SCR_HEIGHT * Y_STEP);


  private boolean[][] scr = null;  
                 // stores if the LEDs are on or off
  private Rectangle[][] scrRects;
       // stores the bounds of each LED
       // used to determine if a mouse click is over a LED


  public DrawPanel()
  {
    setBackground(Color.BLACK);
    scr = new boolean[SCR_HEIGHT][SCR_WIDTH];

    // calculate the bounds of each LED
    scrRects = new Rectangle[SCR_HEIGHT][SCR_WIDTH];
    int y = EDGE_OFFSET;
    for (int r = 0; r < SCR_HEIGHT; r++) {
      int x = EDGE_OFFSET;
      for (int c = 0; c < SCR_WIDTH; c++) {
        scrRects[r][c] = new Rectangle(x, y, LIGHT_WIDTH, LIGHT_HEIGHT);
        x += X_STEP;
      }
      y += Y_STEP;
    }

    // toggle a light using a mouse press or drag
    addMouseListener(new MouseAdapter() {
      public void mousePressed(MouseEvent e)
      { toggleLED(e.getX(), e.getY());
        repaint();
      }
    });

  } // end of DrawPanel()


  private void toggleLED(int x, int y)
  // toggle the LED at (x,y)
  { for (int r = 0; r < SCR_HEIGHT; r++)
      for (int c = 0; c < SCR_WIDTH; c++)
        if (scrRects[r][c].contains(x,y)) {
          scr[r][c] = !scr[r][c];
          break;
        }
  }  // end of toggleLED()


  public void clear()
  // turn off all the LEDs
  { for (int r = 0; r < SCR_HEIGHT; r++)
      for (int c = 0; c < SCR_WIDTH; c++)
        scr[r][c] = false;
  }  // end of clear()



  public void invert()
  // invert all the LEDs: on <-> off
  { for (int r = 0; r < SCR_HEIGHT; r++)
      for (int c = 0; c < SCR_WIDTH; c++)
        scr[r][c] = !scr[r][c];
  }  // end of clear()



  public void printSprite()
  // print the current state of the LEDs to stdout
  {
    System.out.println("-------- " + SCR_WIDTH + "x" + 
                                     SCR_HEIGHT + " -------");
    for (int r = 0; r < SCR_HEIGHT; r++) {
      for (int c = 0; c < SCR_WIDTH; c++)
        System.out.print( (scr[r][c]) ? "*" : " ");
      System.out.println();
    }
    System.out.println("---------------------");
  }  // end of printSprite()



  // ----- redefined methods inherited from JPanel ---------------


  public Dimension getPreferredSize()
  {   return new Dimension(BOARD_WIDTH, BOARD_HEIGHT); }


  public void paintComponent(Graphics s)
  // repaint the panel by drawing all the LED
  { 
    super.paintComponent(s);   // repaint standard stuff first
    Graphics2D g2 = (Graphics2D) s;

    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                        RenderingHints.VALUE_ANTIALIAS_ON);
    g2.setColor(Color.BLACK);
    g2.fillRect(0, 0, BOARD_WIDTH, BOARD_HEIGHT);

    for (int r = 0; r < SCR_HEIGHT; r++) {
      for (int c = 0; c < SCR_WIDTH; c++) {
        if (scr[r][c]) {   // LED is "on"
          litLED(g2, scrRects[r][c].x, scrRects[r][c].y, 
                                        LIGHT_WIDTH, LIGHT_HEIGHT);
        }
        else {   // LED is "off"
          g2.setColor(Color.GRAY);
          g2.drawRect(scrRects[r][c].x, scrRects[r][c].y, 
                                        LIGHT_WIDTH, LIGHT_HEIGHT); 
        }
      }
    }
  }  // end of paintComponent()



  private void litLED(Graphics2D g2, int x, int y, int w, int h)
  /* an "on" LED is draw as a rectangle filled with a radial gradient.
     The LED is located at (x,y) with dimensions w x h
  */
  {
    Point2D center = new Point2D.Float((float)x + w/2.0f, (float)y + h/2.0f);
    float radius = h/2.0f;
    float[] dist = {0.4f, 0.6f};
    Color[] colors = {Color.ORANGE, Color.RED};
    RadialGradientPaint paint = 
               new RadialGradientPaint(center, radius, dist, colors);
    g2.setPaint(paint);

    g2.fillRect(x, y, w, h); 
  }  // end of litLED()


} // end of DrawPanel class
