
// HIDScreen.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th

/* Uses a fork of the javahidapi library at:  
      https://github.com/torbjornvatn/javahidapi

  This makes it easy to communicate with the 
  Dream Cheeky LED Message Board, which is a USB HID device.

  The board in a 2D array of LEDs: 21 across, 7 down.
  The x-axis runs across, as rows 0 to 20.
  The y-axis runs down, as columns 0 to 6.

  Much of my HID code is based on:
    https://charliex2.wordpress.com/tag/dream-cheeky/

  The board accepts a 9-byte command, with the format:
    byte 0: always 0
    byte 1: brightness level, can be 0-15
    byte 2: row number (in multiples of 2: 0, 2, 4, 6)
    
    The rest of the bytes affect two rows of LEDs starting from
    the specified row number.
    
    First row:
      byte 5: 0x00 – left 8 LEDs
      byte 4: 0x00 – middle 8 LEDs
      byte 3: 0x00 – right hand side 5 LEDs

    Second row:
      byte 8: 0x00 – left 8 LEDs
      byte 7: 0x00 – middle 8 LEDs
      byte 6: 0x00 – right hand side 5 LEDs

  An hexadecimal is read as 8 bits, and applied to the board's 
  LEDs left to right by reading the bits right to left!

  0 == LED light on; 1 == light off

  Examples:
  * 0xfe (1111 1110) in byte 8 refers to the second
    row's left group of 8 LEDs. 
    Reading the bits right to left means that the 
    leftmost LED is on and the others off.

  * 0xaa  (1010 1010) 
    Reading the bits right to left means that the 
    LEDs left to right are:
        ON, OFF, ON,OFF, ON,OFF, ON, OFF
    (the LEDs are reversed, and 0 is ON)

*/

import java.io.*;
import java.awt.*;
import java.util.*;

import com.codeminders.hidapi.*;



public class HIDScreen extends Thread implements Board
{
  private static boolean isAvailable = false;  // is the board available?

  // load the native HID library
  static {
    try {
      System.loadLibrary("hidapi-jni");  
          // renamed on Windows from "hidapi-jni-32" or "hidapi-jni-64"
      isAvailable = true;
    }
    catch (UnsatisfiedLinkError e) {
      // System.out.println("Could not find HID library");
      isAvailable = false;
    }
  }


  // Dream Cheeky LED Message Board constants
  private static final int VENDOR_ID = 0x1D34;
  private static final int PRODUCT_ID = 0x0013;

  private static final int CMD_LEN = 9;  
             // length of byte array for the command

  private static final int REFRESH_DELAY = 200;  // ms
       // interval between each refresh of the board's LEDs

  // number of LEDS across and down the board 
  private static final int SCR_WIDTH = 21;
  private static final int SCR_HEIGHT = 7;


  private HIDManager hidManager = null;
  private HIDDevice dev = null;

  private volatile boolean isRunning = false;   
                      // used to stop the refresh thread
  private boolean[][] ledArr = null;   
     // represents the board's array of LEDs in row x column order;
     // true means a light is ON; false means off



  public HIDScreen()
  {
    ledArr = new boolean[SCR_HEIGHT][SCR_WIDTH];
    for (int r = 0; r < SCR_HEIGHT; r++)
       for(int c = 0; c < SCR_WIDTH; c++)
         ledArr[r][c] = false;   // all LEDs are off

    System.out.println("HID Library " +
              ((isAvailable)? "available" : "unavailable"));
    if (isAvailable) {
      dev = openDevice(VENDOR_ID, PRODUCT_ID);
      if (dev != null)
        this.start();
    }
  }  // end of HIDScreen



  public void run()
  /* run() contains a loop which keeps refreshing the board's
     LEDs every REFRESH_DELAY ms based on the contents of the
     ledArr 2D array.
  */
  {
    isRunning = true;
    while(isRunning) {
      refreshBoard(dev);
      try {
        Thread.sleep(REFRESH_DELAY);
      }
      catch(InterruptedException ex){}
    }
    release(dev);
  } // end of run()


  public void draw(char[][] scr)
  {
    int width = scr[0].length;
    int height = scr.length;
    if ((width != SCR_WIDTH) || (height != SCR_HEIGHT)) 
      System.out.println("Supplied draw() array is incorrect size");
    else {
      for (int r = 0; r < SCR_HEIGHT; r++)
        for (int c = 0; c < SCR_WIDTH; c++)
          ledArr[r][c] = (scr[r][c] == '*')? true : false;
    }
  }  // end of draw()



  public void close()
  {  isRunning = false;   }


  public boolean isAvailable()
  {  return isAvailable;  }


  public Dimension getBoardSize()
  {  return new Dimension(SCR_WIDTH, SCR_HEIGHT);  }


  // method for stand-alone testing via main() calls
  public void set(int x, int y, boolean isOn)
  {  
    if ((x < 0) || (x >= SCR_WIDTH) ||
        (y < 0) || (y >= SCR_HEIGHT))
      System.out.println("Light out of range");
    else
      ledArr[y][x] = isOn;
  } // end of set()



  // ------ HID methods -------------

  private HIDDevice openDevice(int vid, int pid)
  /* open a link to the HID device with the supplied vendor and
     product IDs.
  */
  {
    HIDDevice dev = null;
    try {
      hidManager = HIDManager.getInstance();
      if (hidManager == null) {
        System.out.println("No HID manager found");
        isAvailable = false;
        return null;
      }

      dev = hidManager.openById(vid, pid, null);
      if (dev == null) {
        System.out.println("No HID found with vendor/product IDs: " +
                                        vid + ", " + pid);
        hidManager.release();
        isAvailable = false;
      }
      else {
        System.out.println("Opened HID device:");
        System.out.println("  Manufacturer: " +
                              dev.getManufacturerString());
        System.out.println("  Product: " +
                              dev.getProductString());
        System.out.println("  Serial Number: " +
                              dev.getSerialNumberString());
        isAvailable = true;
      }
    }
    catch (Exception e) {
      System.out.println("Could not open HID device");
      isAvailable = false;
    }
    return dev;
  }  // end of openDevice()



  private void release(HIDDevice dev)
  // close the link to the HID device
  {
    try {
      dev.close();
      hidManager.release();
      System.out.println("Released HID device");
    }
    catch (IOException e) {
      System.out.println("Could not release HID");
    }
  }  // end of release()



  private void refreshBoard(HIDDevice dev)
  /* Use the boolean array to build a byte array
     of the form described in the comments at the top
     of this class, and then send it to the device.
  */
  {
    if (ledArr == null)
      return;

    byte[] cmd = new byte[CMD_LEN];
    int r = 0;  // row number
    while(r < SCR_HEIGHT) {
      cmd[0] = (byte)0;  // always 0
      cmd[1] = (byte)1;  // brightness 0 .. 15
      cmd[2] = (byte)r;  // row number (in multiples of 2: 0, 2, 4, 6)

      // first row
      // left 8 LEDs
      cmd[5] = led8( at(ledArr,r,0), at(ledArr,r,1),
                     at(ledArr,r,2), at(ledArr,r,3),
                     at(ledArr,r,4), at(ledArr,r,5),
                     at(ledArr,r,6), at(ledArr,r,7) );
      // middle 8 LEDs
      cmd[4] = led8( at(ledArr,r,8), at(ledArr,r,9),
                     at(ledArr,r,10), at(ledArr,r,11),
                     at(ledArr,r,12), at(ledArr,r,13),
                     at(ledArr,r,14), at(ledArr,r,15) );
      // right 5 LEDs (last 3 bytes are ignored)
      cmd[3] = led8( at(ledArr,r,16), at(ledArr,r,17), 
                     at(ledArr,r,18), at(ledArr,r,19),
                     at(ledArr,r,20), (byte)0, (byte)0, (byte)0);

      // second row
      r++;
      if (r < SCR_HEIGHT) { //stops an attempt to fill non-existent row 7
        // left 8 LEDs
        cmd[8] = led8( at(ledArr,r,0), at(ledArr,r,1), 
                       at(ledArr,r,2), at(ledArr,r,3),
                       at(ledArr,r,4), at(ledArr,r,5), 
                       at(ledArr,r,6), at(ledArr,r,7) );
        // middle 8 LEDs
        cmd[7] = led8( at(ledArr,r,8), at(ledArr,r,9), 
                       at(ledArr,r,10), at(ledArr,r,11),
                       at(ledArr,r,12), at(ledArr,r,13), 
                       at(ledArr,r,14), at(ledArr,r,15) );
        // right 5 LEDs (last 3 bytes are ignored)
        cmd[6] = led8( at(ledArr,r,16), at(ledArr,r,17), 
                       at(ledArr,r,18), at(ledArr,r,19),
                       at(ledArr,r,20), (byte)0, (byte)0, (byte)0);
        r++;
      }

      try {
        dev.write(cmd);
      }
      catch (Exception e) {
        System.out.println("Write error: " + e);
      }
    }
  }  // end of refreshBoard()



  private byte at(boolean[][] ledArr, int r, int c)
  /* convert the boolean at ledArr[r][c] into a byte.
           1 == LED light on; 0 == light off
     but the values are negated by led8() 
  */
  { return (ledArr[r][c]) ? (byte)1 : (byte)0;  }



  private byte led8(byte a, byte b, byte c, byte d,
                           byte e, byte f, byte g, byte h)
  /* The vars a to h represents LED lights in left to right order.
     The constructed hexadecimal stores them as bits in right to left
     order, and negates the bits since a 0 bit == LED light on,
     a 1 bit == light off 
  */
  {  return (byte) (~(h<<7) & ~(g<<6) & ~(f<<5) & ~(e<<4) &
                    ~(d<<3) & ~(c<<2) & ~(b<<1) & ~(a));  }




  public static void listDevices()
  // print details of all the HID devices
  {
    try {
      HIDManager hm = HIDManager.getInstance();
      HIDDeviceInfo[] devs = hm.listDevices();

      System.out.println("HID Devices:");
      for (int i = 0; i < devs.length; i++)
        System.out.println("" + i + ". " + devs[i] + "\n");
      System.out.println();
    }
    catch (Exception e) {
      System.out.println("  Unable to list devices: " + e);
    }
  }  // end of listDevices()


  // ------------------ test rig -----------------


  public static void main(String[] args)
  { 
    if (!isAvailable) {
      System.out.println("Could not find HID library");
      return;
    }

    listDevices();

    HIDScreen scr = new HIDScreen();
    if (scr.isAvailable()) {
      scr.set(5, 5, true);
      scr.set(6, 5, true);
      scr.set(7, 5, true);
      scr.set(7, 4, true);
      scr.set(7, 3, true);
      
      Utils.delay(2000);
      scr.close();
    }
  }   // end of main()


}  // end of HIDScreen class

