
// LedPanel.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th

/* A JPanel acting as a canvas for drawing a version of
   the Dream Cheeky LED message board.

   Each LED is drawn as a rectangle.

   This class is quite similar to DrawPanel which is used
   for drawing sprites.
*/

import java.awt.*;
import javax.swing.*;
import java.awt.geom.*;


public class LedPanel extends JPanel
{
  // dimensions of the screen in terms of number of LED lights
  private static final int SCR_WIDTH = 21;
  private static final int SCR_HEIGHT = 7;

  // pixel dimensions of a LED light 
  private static final int LIGHT_WIDTH = 10;
  private static final int LIGHT_HEIGHT = 7;

  // distance between each LED
  private static final int X_OFFSET = 5;
  private static final int Y_OFFSET = 5;   // same spacing

  // distance between LEDs and the pane's edges
  private static final int EDGE_OFFSET = 10;

  // distance from one LED to the next
  private static final int X_STEP = LIGHT_WIDTH + X_OFFSET;
  private static final int Y_STEP = LIGHT_HEIGHT + Y_OFFSET;

  // pixel dimensions of the pane
  private static final int BOARD_WIDTH = 
                              EDGE_OFFSET + (SCR_WIDTH * X_STEP);
  private static final int BOARD_HEIGHT = 
                              EDGE_OFFSET + (SCR_HEIGHT * Y_STEP);


  private boolean[][] ledArr = null;
                 // stores if the LEDs are on or off


  public LedPanel()
  {
    setBackground(Color.BLACK);
    ledArr = new boolean[SCR_HEIGHT][SCR_WIDTH];
    for (int r = 0; r < SCR_HEIGHT; r++)
      for (int c = 0; c < SCR_WIDTH; c++)
        ledArr[r][c] = false;  // all the LEDs are off 
  }  // end of LedPanel()



  public void draw(char[][] scr)
  // use the screen char array to set the booleans in ledArr[][]
  {
    int width = scr[0].length;
    int height = scr.length;
    if ((width != SCR_WIDTH) || (height != SCR_HEIGHT)) 
      System.out.println("Supplied draw() array is incorrect size");
    else {
      for (int r = 0; r < SCR_HEIGHT; r++)
        for (int c = 0; c < SCR_WIDTH; c++)
          ledArr[r][c] = (scr[r][c] == '*')? true : false;
    }
    repaint();  // trigger redrawing
  }  // end of draw()


  public Dimension getBoardSize()
  {  return new Dimension(SCR_WIDTH, SCR_HEIGHT);  }



  public void set(int x, int y, boolean isOn)
  // turn the LED at ledArr[y][x] on or off
  {  
    if ((x < 0) || (x >= SCR_WIDTH) ||
        (y < 0) || (y >= SCR_HEIGHT))
      System.out.println("Light out of range");
    else
      ledArr[y][x] = isOn;
    repaint();  // trigger redrawing
  } // end of set()




  // ----- redefined methods inherited from JPanel ---------------


  public Dimension getPreferredSize()
  {   return new Dimension(BOARD_WIDTH, BOARD_HEIGHT); }



  public void paintComponent(Graphics g)
  // repaint the panel by drawing all the LEDs
  {
    super.paintComponent(g);  // repaint standard stuff first

    Graphics2D g2 = (Graphics2D)g;
    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                        RenderingHints.VALUE_ANTIALIAS_ON);
    g2.setColor(Color.BLACK);
    g2.fillRect(0, 0, BOARD_WIDTH, BOARD_HEIGHT);

    int y = EDGE_OFFSET;
    for (int r = 0; r < SCR_HEIGHT; r++) {
      int x = EDGE_OFFSET;
      for (int c = 0; c < SCR_WIDTH; c++) {
        if (ledArr[r][c])    // LED is "on"
          litLED(g2, x, y, LIGHT_WIDTH, LIGHT_HEIGHT);
        else {  // LED is "off"
          g2.setColor(Color.GRAY);
          g2.drawRect(x, y, LIGHT_WIDTH, LIGHT_HEIGHT); 
        }
        x += X_STEP;
      }
      y += Y_STEP;
    }
  }  // end of paintComponent()



  private void litLED(Graphics2D g2, int x, int y, int w, int h)
  /* an "on" LED is draw as a rectangle filled with a radial gradient.
     The LED is located at (x,y) with dimensions w x h
  */
  {
    Point2D center = new Point2D.Float((float)x + w/2.0f, (float)y + h/2.0f);
    float radius = h/2.0f;
    float[] dist = {0.4f, 0.6f};
    Color[] colors = {Color.ORANGE, Color.RED};
    RadialGradientPaint paint = 
               new RadialGradientPaint(center, radius, dist, colors);
    g2.setPaint(paint);

    g2.fillRect(x, y, w, h); 
  }  // end of litLED()


}  // end of LedPanel class
