
// Screen.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th


/*  This class deals with sprites at any position on the screen,
    but stores the ones within the screen rectangle as characters
    in a scr[][] char array. '*' means the pixel is on, ' ' means off

    The top-left corner of the screen rectangle is at (0,0) and is
    <sz.width> units wide, and <sz.height> deep. 
    The + x-axis runs across, the + y-axis runs down. 

    Each screen pixel (character in the scr[][] array) is mapped to a LED
    light in the HID screen or to a drawn rectangle in the Swing-based
    LedScreen.

    LedScreen is only used if an instance of the HIDScreen cannot be
    created, probably because the Dream Cheeky LED Message Board is not
    plugged into the machine.

    Main method categories:
         * screen info
         * screen testing: for intersection, if point is on the screen
         * screen clearing
         * adding sprites, char arrays, points
         * adding shapes: line, rectangles, circles

*/

import java.io.*;
import java.awt.*;
import java.util.*;



public class Screen
{
  // refers to the HID or Swing-based screen
  private Board board = null;
  private Dimension sz;  // size of the board

  private char[][] scr;
      // a char array representing the screen;
      // '*' means the pixel is on, ' ' means off

  private Rectangle scrRect;  // used for intersection testing



  public Screen()
  {
    // create the HID screen or a Swing-based screen
    board = new HIDScreen();
    if (!board.isAvailable())
      board = new LedScreen();  // always available

    sz = board.getBoardSize();
    scr = new char[sz.height][sz.width];
    scrRect = new Rectangle(0, 0, sz.width, sz.height);
  }  // end of Screen()



  public void draw()
  // pass char array to the 'real' screen for rendering
  {   board.draw(scr);  } 



  public void close()
  { board.close();  }


  // ---------------- screen info ---------------


  public int getWidth()
  {  return board.getBoardSize().width;  }


  public int getHeight()
  {  return board.getBoardSize().height;  }


  public void print()
  {  ArrUtils.print(scr);  }



  // ---------------- screen testing ---------------

  public Rectangle intersection(Sprite s)
  // how does s intersect with the screen
  {  return scrRect.intersection(s.getRect());  } 


  public boolean intersects(Sprite s)
  // does s intersect with the screen?
  {  return scrRect.intersects(s.getRect());  } 


  public boolean onScreen(Point p)
  {  return ArrUtils.inside(scr, p);  }


  public boolean onScreen(int x, int y)
  {  return ArrUtils.inside(scr, x, y);  }



  // -------- screen clearing ----------------

  public void clear()
  // wipe the screen clear
  {  ArrUtils.clear(scr);  }


  public void clear(int x, int y, int w,  int h)
  {  ArrUtils.clear(scr, x, y, w, h);  }


  public void clear(Rectangle r)
  {  ArrUtils.clear(scr, r);  }


  public void clear(Sprite s)
  // clear the rectangle containing s
  {  ArrUtils.clear(scr, scrRect.intersection(s.getRect()) );  } 



  // --------- adding sprites, arrays, points ----------


  public boolean add(Sprite s)
  // add sprite s to the screen array
  {
    // extract the part of the sprite that overlaps the screen
    Rectangle r = intersection(s);
    char[][] chArr = s.getRegion(r);
    if (chArr == null) {
      // System.out.println("Region is null");
      return false;
    }
    else {
      // ArrUtils.print(chArr);
      return ArrUtils.add(scr, r.x, r.y, chArr);
    }
  }  // end of add()



  public boolean add(int x, int y, char[][] chArr)
  /* add the char array to the screen with its top-left
     corner at (x,y) */
  {  return ArrUtils.add(scr, x, y, chArr);  }




  public void set(int x, int y, boolean isOn)
  // turn scr[y][x] on or off
  { if (ArrUtils.inside(scr, x, y))
      scr[y][x] = (isOn) ? '*' : ' ';
  }  


  public boolean isSet(int x, int y)
  // is scr[y][x] on?
  { 
    if (ArrUtils.inside(scr, x, y))
      return (scr[y][x] == '*');
    return false;
  }  



  // ------------ adding shapes ------------------


  public void addLine(int x0, int y0, int x1, int y1)
  /* add a line to the screen array running from (x0,y0) 
     to (x1,y1)  */
  {
    ArrayList<Point> pts = Utils.calcLine(x0,y0, x1,y1);
         // add the line as a list of points
    for (Point pt : pts)
      if (ArrUtils.inside(scr, pt))
        scr[pt.y][pt.x] = '*';
  }  // end of drawLine()



  public void addRect(int x, int y, int w, int h)
  /* add a rectangle to the screen array which is (w x h) in
     size and has its top-left point at (x,y)
  */
  {
    ArrayList<Point> pts = Utils.calcRect(x, y, w, h);
         // add the rectangle as a list of points
    for (Point pt : pts)
      if (ArrUtils.inside(scr, pt))
        scr[pt.y][pt.x] = '*';
  }  // end of drawRect()




  public void addCircle(int x, int y, int radius)
  /* add a circle to the screen array centered at (x,y) 
     with radius  */
  {
    ArrayList<Point> pts = Utils.calcCircle(x,y, radius);
         // add the circle as a list of points
    for (Point pt : pts)
      if (ArrUtils.inside(scr, pt))
        scr[pt.y][pt.x] = '*';
  }  // end of drawCircle()



  }  // end of Screen class

