
// Sprite.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th

/*  A simple spite consists of a char array of *'s and spaces, 
    called a picture, a name for the picture, and a position for 
    the picture's top-left corner in screen coordinates.

    In fact, a sprite may contain a list of pictures and names,
    with an integer indicating which is the current one. This allows
    a sprite to change its displayed picture.

    Main method categories:
         * position methods
         * picture (array) methods

*/

import java.awt.*;
import java.util.*;


public class Sprite
{
  // specify picture coordinates as compass names
  public static final int NORTH = 0;
  public static final int NE = 1;
  public static final int EAST = 2;
  public static final int SE = 3;
  public static final int SOUTH = 4;
  public static final int SW = 5;
  public static final int WEST = 6;
  public static final int NW = 7;
  public static final int CENTER = 8;


  private static final String DEFAULT_NAME = "noname$$";

  private ArrayList<String> picNames;
  private ArrayList<char[][]> picList;
  private int currPic;   
          // which picture (and name) is the current one displayed
  private Point pos;   // of the top-left corner of the picture

  private char[][] backup =  null;   
                // stores a copy of the charArr as a backup


  public Sprite(char[][] charArr)
  {  createPic(DEFAULT_NAME, charArr);  } 


  public Sprite(String name, char[][] charArr)
  {  createPic(name, charArr);   } 


  private void createPic(String name, char[][] charArr)
  {
    if (charArr == null) {
      System.out.println("Picture is null");
      return;
    }
    if (name == null) {
      System.out.println("No name supplied");
      name = DEFAULT_NAME;
    }

    picNames = new ArrayList<String>();
    picNames.add(name);

    picList = new ArrayList<char[][]>();
    picList.add(ArrUtils.copy(charArr));

    currPic = 0;
    pos = new Point(0,0);
    backup = ArrUtils.copy(charArr);
  }  // end of createPic()



  public Sprite(ArrayList<char[][]> charArrs)
  {
    if (charArrs == null) {
      System.out.println("Pics List is null");
      return;
    }
    if (charArrs.size() == 0) {
      System.out.println("Pics list is empty");
      return;
    }

    createPic(DEFAULT_NAME + "0", charArrs.get(0));
    for(int i=1; i < charArrs.size(); i++)
      addPic(DEFAULT_NAME + "i", charArrs.get(i));
  }  // end of Sprite()



  public Sprite(ArrayList<String> names, ArrayList<char[][]> charArrs)
  {
    if ((names == null) || (charArrs == null)) {
      System.out.println("Null argument");
      return;
    }
    if (charArrs.size() == 0) {
      System.out.println("Pics list is empty");
      return;
    }

    if (names.size() != charArrs.size())
      System.out.println("Names and arrays lists are not the same length");

    createPic(names.get(0), charArrs.get(0));
    for(int i=1; i < charArrs.size(); i++) {
      if (i < names.size())
        addPic(names.get(i), charArrs.get(i));
      else
        addPic(DEFAULT_NAME, charArrs.get(i));
    }
  }  // end of Sprite()



  public void print()
  {
    int np = numPics();
    System.out.println("Pos: (" + pos.x + ", " + pos.y +")");
    if (np == 1) {
      System.out.println("Pic nm: \"" + picNames.get(0) + "\"");
      ArrUtils.print(picList.get(0));
    }
    else {
      System.out.println("Contains " + np + " pictures --");
      for(int i=0; i < np; i++) {
        System.out.println("Pic nm: \"" + picNames.get(i) + "\" (" + i + ")" +
                           ((currPic == i) ? "*:" : ":"));
        ArrUtils.print(picList.get(i));
      }
    }
  }  // end of print()



  // ---------- position methods ------------------

  public int getX()
  {  return pos.x; }

  public int getY()
  {  return pos.y;  }

  public Point getPos()
  {  return pos; }

  
  public Rectangle getRect()
  {  return new Rectangle(pos.x, pos.y, getWidth(), getHeight());  }


  public void setX(int x)
  {  pos.x = x; }

  public void setY(int y)
  {  pos.y = y;  }

  public void setPos(int x, int y)
  {  pos.x = x;  pos.y = y;  }

  public void setPos(Point p)
  {  pos.x = p.x;  pos.y = p.y;  }


  public void setPos(int com, int x, int y)
  /* set the char array's compass point to be at (x,y)
     which means calculating the position of the array's 
     top-left point relative to (x,y), and storing that in pos
  */
  {
    int width = getWidth();
    int height = getHeight();
    int xLeft = x;   // assuming com is NW
    int yTop = y;

    if (com == NORTH)
      xLeft -= width/2;
    else if (com == NE)
      xLeft -= width; 
    else if (com == EAST) {
      xLeft -= width;
      yTop -= height/2; 
    }
    else if (com == SE) {
      xLeft -= width;
      yTop -= height; 
    }
    else if (com == SOUTH) {
      xLeft -= width/2;
      yTop -= height; 
    }
    else if (com == SW)
      yTop -= height; 
    else if (com == WEST) 
      yTop -= height/2; 
    else if (com == NW)  {}  // no offset needed
    else if (com == CENTER) {
      xLeft -= width/2;
      yTop -= height/2; 
    }
    else 
      System.out.println("Unknown compass point; using NW");

    pos.x = xLeft;  
    pos.y = yTop;
  }  // end of setPos()




  public void xStep(int xStep)
  {  pos.x += xStep; }

  public void yStep(int yStep)
  {  pos.y += yStep;  }



  // ---------- picture (array) methods ------------------------



  public void addPic(String name, char[][] charArr)
  {  
    picNames.add(name);
    picList.add(charArr);
  } 


  public int getCurrPicNo()
  {  return currPic;  }


  public int numPics()
  {  return picNames.size();  }


  public String getNames()
  { 
    if (picNames.size() == 1)
      return getName();
    else {
      StringBuilder sb = new StringBuilder();
      for (String nm : picNames)
        sb.append(nm + " ");
      return sb.toString().trim();
    }
  }  // end of getNames()



  public String getName()
  {  return getName(currPic);  }


  public String getName(int i)
  {  
    if ((i < 0) || (i > picNames.size()-1)) {
      System.out.println("Pic name out of range; returning first");
      return picNames.get(0);
    }
    else
      return picNames.get(i); 
  }  // end of getName()



  public int getWidth()
  {  
    char[][] charArr = getCharArray(currPic);
    if (charArr == null)
      return 0;
    else
      return charArr[0].length;
  }  // end of getWidth()


  public int getHeight()
  {  
    char[][] charArr = getCharArray(currPic);
    if (charArr == null)
      return 0;
    else
      return charArr.length;
  }  // end of getHeight()




  public char[][] getCharArray()
  {  return getCharArray(currPic);  }


  public char[][] getCharArray(int i)
  {  
    if ((i < 0) || (i > picList.size()-1)) {
      System.out.println("Pic number out of range; returning first");
      return ArrUtils.copy(picList.get(0));
    }
    else
      return ArrUtils.copy(picList.get(i)); 
  }  // end of getCharArray()



  public void backup()
  {  backup = getCharArray(currPic);  } 


  public void restore()
  {  
   if (backup == null)
     System.out.println("No backup to restore");
   else
     picList.set(currPic, backup);
  }



  public void appendPic(Sprite s)
  /* append picture (char array) in s to the end of this sprite's 
     current picture (char array), with a blank column between them
  */
  { // s.print();
    int width = getWidth();
    int height = getHeight();

    int nWidth = width + s.getWidth() + 1;  // 1 for blank column
    // System.out.println("appendPic nWidth: " + nWidth);
    int nHeight = (height > s.getHeight()) ? height : 
                                             s.getHeight();

    // copy sprite's char array and s' char array into a new array
    char[][] pic = new char[nHeight][nWidth];
    ArrUtils.clear(pic);
    ArrUtils.append(pic, 0, getCharArray());
    ArrUtils.append(pic, width+1, s.getCharArray());  // 1 for blank

    // update the sprite's current picture
    picList.set(currPic, pic);
  }  // end of appendPic();



  public void trim()
  /* remove the starting and ending blank columns from the 
     sprite's current picture (char array)
  */
  { char[][] pic = ArrUtils.trim(getCharArray());
    picList.set(currPic, pic);
  }  // end of trim()



  public char[][] getRegion(Rectangle rect)
  /* return a copy of the char array covered by the 
     rectangle, or null */
  {  return ArrUtils.copyRegion(rect, pos, getCharArray()); } 



  public void transform(String methodNm)
  /* Transform the current picture, creating a new char array.
     The transformation names are:
       copy, invert, rotRight, rotLeft, flipVert, flipHoriz
  */
  { char[][] ca = ArrUtils.transform(methodNm, getCharArray());
    picList.set(currPic, ca);
  } 



  // ------------------ for multiple pics in a sprite ------------


  public boolean toPic(String nm)
  // change the current picture based on its name
  {
    // find the index position of the name
    int pos = -1;
    for (int i=0; i < picNames.size(); i++)
      if (nm.equals(picNames.get(i))) {
        pos = i;
        break;
      }

    if (pos == -1) {
      System.out.println("Pic: " + nm + " not found");
      return false;
    }
    else
      return toPic(pos);
  }  // end of toPic();



  public boolean toPic(int i)
  // change the current picture based on its index position
  {
    if ((i < 0) || (i > picList.size()-1)) {
      System.out.println("Pic index is out of range");
      return false;
    }
    else {
      currPic = i;
      // System.out.println("Switched to pic: " + currPic);
      return true;
    }
  }  // end of toPic()



  public boolean nextPic()
  // change to the next picture, if possible
  {
    if ((currPic == picList.size()-1)) {
      System.out.println("No next picture");
      return false;
    }
    else {
      currPic++;
      // System.out.println("Switched to pic: " + currPic);
      return true;
    }
  }  // end of nextPic()


}  // end of Sprite class

