
// SpriteStore.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th


/* The sprite store maintains two maps of name-sprites, one for the
   characters of a loaded font, and the other for loaded sprites.

   Main method categories:
     * print maps info
     * retrieve a sprite
     * sprite loading
       - load font sprites from a file into the font map
          - loadFont()
       - load sprites from a file into the sprites map
          - loadSprites()
       - load multi-picture sprite from a file into the sprites map
          - loadMultiSprite()
*/

import java.io.*;
import java.awt.*;
import java.util.*;




public class SpriteStore
{
  // font names, reflecting their relative sizes
  public static final int BIG = 0;
  public static final int MEDIUM = 1;
  public static final int SMALL = 2;

  // size of a char array for a sprite pic
  private static final int CHAR_WIDTH = 5;
  private static final int CHAR_HEIGHT = 7;    
                               // same as DC msg board height


  private static SortedMap<String, Sprite> fontMap;
      // used to store the sprites representing the font characters

  private static SortedMap<String, Sprite> spritesMap;
     /* used to store user-defined sprites, loaded with calls to
        loadSprites() and loadMultiSprite()  
     */


  public SpriteStore()
  {  this(BIG);  }


  public SpriteStore(int fontType)
  {
    // System.out.println("Loading font type: " + fontType);
    fontMap = new TreeMap<String, Sprite>();
    spritesMap = new TreeMap<String, Sprite>();

    // load one of the fonts
    if (fontType == BIG)
      loadFont("bigFont.txt", true);
    else if (fontType == MEDIUM)
      loadFont("mediumFont.txt", false);
    else if (fontType == SMALL)
      loadFont("smallFont.txt", false);
    else {
      System.out.println("Font type not recognized; loading big font");
      loadFont("bigFont.txt", true);
    }
  }  // end of SpriteStore()


  public SpriteStore(String fnm)
  {
    // System.out.println("Loading font: " + fnm);
    fontMap = new TreeMap<String, Sprite>();
    spritesMap = new TreeMap<String, Sprite>();
    loadFont(fnm, false);
  }  // end of SpriteStore()


  // ------------- print maps info -------------------


  public void printNames()
  { printNames("Font", fontMap);
    printNames("Sprites", spritesMap);
  }



  private void printNames(String mapNm, SortedMap<String, Sprite> map)
  // print the names of all the sprites stored in map
  {
    System.out.println(mapNm + ":");
    int i = 1;
    for (String nm: map.keySet()){
      int numPics = getSprite(map, nm).numPics();
      if (numPics > 1)
        System.out.print(nm + "(" + numPics + ") "); 
      else
        System.out.print(nm + " ");
      if (i++ % 6 == 0)
        System.out.println();
    }
    System.out.println("\n");
  }  // end of print()



  public void print()
  { print("Font", fontMap);
    print("Sprites", spritesMap);
  }


  private void print(String mapNm, SortedMap<String, Sprite> map)
  // print all the sprites' names and pictures stored in map
  {
    System.out.println(mapNm + ":");
    for (String nm: map.keySet()){
      System.out.println(nm + ":"); 
      Sprite s = getSprite(map, nm);
      if (s != null)
        s.print();
    }
  }  // end of print()



  // ----- retrieve a sprite from one of the maps -----------------


  public Sprite getSprite(String nm)
  {  return getSprite(spritesMap, nm);  }


  public Sprite getSprite(char ch)
  // get the sprite for a character from the font map
  {
    String nm;
    if (ch == ' ')
      nm = "sp";   // special case
    else
      nm = "" + ch;
    return getSprite(fontMap, nm);
  }  // end of getSprite()



  private Sprite getSprite(SortedMap<String, Sprite> map, String nm)
  {
    Sprite s = (Sprite) map.get(nm);
    if (s == null) {
      System.out.println("No sprite found for \"" + nm + "\"");
      return null;
    }
    else
      return s;
  }  // end of getSprite()



  public Sprite buildMsg(String msg)
  /* create a sprite made by combining the pictures for all the 
     characters in the supplied message
  */
  {
    if (msg == null) {
      System.out.println("No message supplied");
      return null;
    }

    Sprite s = getSprite(msg.charAt(0));
    Sprite msgSprite = new Sprite(msg, s.getCharArray());

    for (int i=1; i < msg.length(); i++)
      msgSprite.appendPic( getSprite(msg.charAt(i)) );

    msgSprite.trim();
    return msgSprite;
  }  // end of buildMsg()



  public Sprite buildMsg(String[] spriteNms)
  /* create a sprite made by combining the pictures for all the 
     sprite names in the supplied array
  */
  {
    if (spriteNms == null) {
      System.out.println("No sprite names supplied");
      return null;
    }

    Sprite s = getSprite(spriteNms[0]);
    Sprite msgSprite = new Sprite(s.getCharArray());

    for (int i=1; i < spriteNms.length; i++)
      msgSprite.appendPic( getSprite(spriteNms[i]) );

    msgSprite.trim();
    return msgSprite;
  }  // end of buildMsg()



  // ------------- load font from a file ------------------
  /* the fonts can either specify the character using a vertical
     format for the bytes of the char arrays, or a horizontal format.
  */


  private void loadFont(String fnm, boolean isVertical)
  { 
    System.out.println("Reading font from " + fnm);
    int count = 0;
    try {
      //InputStream is  = this.getClass().getClassLoader().
      //                                  getResourceAsStream("/fonts/"+fnm);
      InputStream is  = this.getClass().getResourceAsStream("/fonts/"+fnm);
      BufferedReader br = new BufferedReader( new InputStreamReader(is));
      String line;
      while((line = br.readLine()) != null) {
        if (line.length() == 0)      // empty line
          continue;
        if (line.startsWith("//"))   // comment
          continue;
        if (isVertical)  // font uses vertical format
          getColumnChar(line);
        else   // horizontal format
          getRowChar(line);
        count++;
      }
      br.close();
    } 
    catch (IOException e) 
    { System.out.println("Error reading characters info file: " + fnm);
    }
    System.out.println("Lines processed: " + count);
  }  // end of loadFont()




  private void getColumnChar(String line)
  /* The 'vertical' format of the input lines are:
          charName  0x7f  0x08  0x04  0x04  0x78   
                    // 5 hexadecimals
     Each hexadecimal represents a *column* in the grid 
     making up the character.
  */
  {
    String[] tokens = line.trim().split("\\s+");
    if (tokens.length != 6)
      System.out.println("Incorrect format for line: \"" + line + "\"");
    else {
      byte[] bytes = new byte[5];
      for(int i=0; i < 5; i++)
        bytes[i] = getByte(tokens[i+1]);
      char[][] chArr = bytesColArray(bytes);
      Sprite s = new Sprite(tokens[0], chArr);
      s.trim();

      fontMap.put(tokens[0], s);
      // System.out.println("Added sprite for \"" + tokens[0] + "\"");
    }
  }  // end of getColumnChar()



  private byte getByte(String s)
  {
    byte val = (byte)0;
    try {
      val = (byte)Integer.decode(s).intValue();
    }
    catch (NumberFormatException e){ 
      System.out.println("Could not extract byte from " + s + ": " + e); 
    }
    return val;
  }  // end of getByte()



  private char[][] bytesColArray(byte[] chBytes)
  /* The array contains 5 hexadecimals.
     Each hexadecimal represents a 7-element column 
     in the grid making up the character.
     The hexadecimal represents the column from 
     top to bottom as 7 bits. 1 means '*', 0 is ' '
     Explained at https://helpfactory.wordpress.com/tag/ascii/
  */
  {
    char [][] charArr = new char[CHAR_HEIGHT][CHAR_WIDTH];
    ArrUtils.clear(charArr);

    for (int col = 0; col < chBytes.length; col++) {
      String bStr = String.format("%7s", 
                       Integer.toBinaryString(chBytes[col])).
                                               replace(' ', '0');
      char[] bits = bStr.toCharArray();
      for(int row = 0; row < bits.length; row++)
        charArr[bits.length-1-row][col] = 
                            (bits[row] == '1') ? '*' : ' ';
    }
    return charArr;
  }  // end of bytesColArray()



  private void getRowChar(String line)
  /* The 'horizontal' format of the input lines are:
          !  0x02  0x02  0x02  0x00  0x02  0x00  0x00   
                    // charname and 7 hexadecimals
     Each hexadecimal represents a *row* in the grid 
     making up the character.
  */
  {
    String[] tokens = line.trim().split("\\s+");
    if (tokens.length != 8)
      System.out.println("Incorrect format for line: \"" + line + "\"");
    else {
      byte[] bytes = new byte[7];
      for(int i=0; i < 7; i++)
        bytes[i] = getByte(tokens[i+1]);
      char[][] charArr = bytesRowArray(bytes);
      Sprite s = new Sprite(tokens[0], charArr);
      s.trim();

      fontMap.put(tokens[0], s);
      // System.out.println("Char \"" + tokens[0] + "\":");
      // ArrUtils.printArray(charArr);
    }
  }  // end of getRowChar()



  private char[][] bytesRowArray(byte[] chBytes)
  /* The array contains 7 hexadecimals.
     Each hexadecimal represents a 5-element row 
     in the grid making up the character.
     The hexadecimal represents the row from 
     right to left as 5 bits (i.e. reversed). 1 means '*', 0 is ' '
  */
  {
    char [][] charArr = new char[CHAR_HEIGHT][CHAR_WIDTH];
    ArrUtils.clear(charArr);

    for (int r = 0; r < CHAR_HEIGHT; r++) {
      String bStr = String.format("%5s", 
                       Integer.toBinaryString(chBytes[r])).
                                               replace(' ', '0');
      // System.out.println(bStr);
      char[] bits = bStr.toCharArray();
      for(int c = 0; c < CHAR_WIDTH; c++)
        charArr[r][c] = (bits[CHAR_WIDTH-1-c] == '1') ? '*' : ' ';
    }
    return charArr;
  }  // end of bytesRowArray()



  // -------- load sprites from a file -------------


  public void loadSprites(String fnm)
  /* The format of the input lines are:
          <sprite name>  <width> <height>   
          // height lines consisting of spaces and *'s up to width columns
     and empty lines and comment lines.
     There may be many sprite name defs in the file.
  */
  { 
    String spriteName = null;
    char[][] chArr = null;
    int rowNo = 0;
    boolean isFillingPic = false;
    System.out.println("Loading sprites from " + fnm);
    try {
      BufferedReader br = new BufferedReader( new FileReader(fnm));
      String line;
      while((line = br.readLine()) != null) {
        if ( (line.length() == 0) &&    // empty line outside of pic
             (!isFillingPic))
          continue;
        if (line.startsWith("//"))   // comment
          continue;

        if (startsWithAlpha(line)) {    // start of sprite info lines
          if (isFillingPic) {    // but still filling pic?
            System.out.println("Missing pic data for \"" + 
                                     spriteName + "\"; not stored");
            isFillingPic = false;
          }

          // start processing next sprite
          String[] tokens = line.trim().split("\\s+");
          if (tokens.length != 3)
            System.out.println("Incorrect format for sprite info: \"" + line + "\"");
          else {
            spriteName = tokens[0].trim();
            Dimension d = getSpriteSize(tokens[1], tokens[2]);
            chArr = new char[d.height][d.width];
            ArrUtils.clear(chArr);
            isFillingPic = true;
            rowNo = 0;
          }
        }
        else if (isFillingPic) {  
          // in the middle of filling a sprite
          fillRow(chArr, rowNo, line);
          rowNo++;
          if (rowNo == chArr.length) {  // finished reading pic data
            Sprite s = new Sprite(spriteName, chArr);
            spritesMap.put(spriteName, s);
            // System.out.println("  Added \"" + spriteName + "\"");
            isFillingPic = false;
          }
        }
      }

      if (isFillingPic)
        System.out.println("Missing pic data for \"" + 
                               spriteName + "\"; not stored");
      br.close();
    } 
    catch (IOException e) 
    { System.out.println("Error reading characters info file: " + fnm);   }

    System.out.println("No. of sprites: " + spritesMap.size());
  }  // end of loadSprites()



  private boolean startsWithAlpha(String line)
  { 
    if ((line == null) || (line.length() == 0))
      return false;
    char ch = Character.toLowerCase(line.charAt(0));
    return ((ch >= 'a') && (ch <= 'z'));
  }



  private Dimension getSpriteSize(String wStr, String hStr)
  {
    int width = getInt(wStr);
    int height = getInt(hStr);
    if ((width <= 0) || (height <= 0)) {
      System.out.println("Sprite dimension incorrect");
      return null;
    }
    else 
      return new Dimension(width, height);
  }  // end of getSpriteSize()



  private int getInt(String s)
  {
    int val = 0;
    try {
      val = Integer.parseInt(s);
    }
    catch (NumberFormatException e){ 
      System.out.println("Could not extract int from " + s); 
    }
    return val;
  }  // end of getInt()



  private void fillRow(char[][] chArr, int rowNo, String line)
  // store the line in the specified row of the char array
  {
    int width = chArr[0].length;

    if (line.length() < width) {
      // System.out.println("Line is short, adding spaces to \"" + line + "\"");
      width = line.length();
    }
    else if (line.length() > width)
      System.out.println("Line too long, only processing first " + 
                                  width + " chars in \"" + line + "\"");

    for (int i=0; i < width; i++) {
      char ch = line.charAt(i);
      if ((ch == '*') || (ch == ' '))
         chArr[rowNo][i] = ch;
      else {
        System.out.println("Unknown char: '" + ch +"'; using '*'");
        chArr[rowNo][i] = '*';
      }
    }
  }  // end of fillRow();




  // ------- load multi-picture sprite --------------------


  public void loadMultiSprite(String fnm)
  /* The multiple picture format for a sprite is:
          <sprite name>  <width> <height>
     and empty lines and comment lines.
     
     Each picture has the form:
          <name of picture>
          // and then lines consisting of spaces and *'s
     and empty lines and comment lines.
  */
  { 
    Sprite sprite = null;
    String spriteName = null;
    String picName = null;
    Dimension d = null;
    char[][] chArr = null;

    int rowNo = 0; 
    boolean hasAnimInfo = false;
    boolean isFillingPic = false;
    boolean isFirstPic = true;

    int numPics = 0;
    System.out.println("Loading multi-pic sprite from " + fnm);
    try {
      BufferedReader br = new BufferedReader( new FileReader(fnm));
      String line;
      while((line = br.readLine()) != null) {
        if ((line.length() == 0) &&    // empty line outside of pic
            (!isFillingPic))
          continue;
        if (line.startsWith("//"))   // comment
          continue;

        if (!hasAnimInfo) {
          if (startsWithAlpha(line)) {   // get animation info
            String[] tokens = line.trim().split("\\s+");
            if (tokens.length != 3)
              System.out.println("Incorrect format for multi-pic info: \"" + 
                                                           line + "\"");
            else {
              spriteName = tokens[0].trim();
              System.out.println("Sprite name: \"" + spriteName + "\"");
              d = getSpriteSize(tokens[1], tokens[2]);
              hasAnimInfo = true;
            }
          }
        }
        else {   // has animation info
          if (startsWithAlpha(line)) {   // start of sprite info line
            if (isFillingPic) {    // but still filling previous pic?
              System.out.println("Missing pic data for \"" + 
                                       picName + "\"; not stored");
              isFillingPic = false;
            }
     
            // start processing next pic
            String[] tokens = line.trim().split("\\s+");
            if (tokens.length != 1)
              System.out.println("Incorrect format for pic info: \"" + line + "\"");
            else {
              picName = tokens[0].trim();
              chArr = new char[d.height][d.width];
              ArrUtils.clear(chArr);
              isFillingPic = true;
              rowNo = 0;
            }
          }
          else {  // in the middle of filling a char array
            fillRow(chArr, rowNo, line);
            rowNo++;
            if (rowNo == chArr.length)  {  // finished reading pic data
              if (isFirstPic) {  // it was the first pic
                sprite = new Sprite(picName, chArr);
                numPics = 1;
                isFirstPic = false;
              }
              else {   // later pic is finished
                sprite.addPic(picName, chArr);
                numPics++;
              }
              System.out.println("  Added pic \"" + picName + "\"");
              isFillingPic = false;
            }
          }
        }
      }

      if (isFillingPic)    // eof but still filling previous pic?
        System.out.println("Missing pic data for \"" + 
                                 picName + "\"; not stored");
      br.close();
    } 
    catch (IOException e) 
    { System.out.println("Error reading characters info file: " + fnm);   }

    if ((sprite != null) && !isFillingPic)
      spritesMap.put(spriteName, sprite);
    System.out.println(numPics + " pics stored in \"" + spriteName + "\"");
  }  // end of loadMultiSprite()


}  // end of SpriteStore class
