
// ToChars.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th

/* Convert an image to a character array.
   Scale the image, convert to black and white, and then
   to an array of '*' (for black) and ' ' (white) characters. 
   Finish by printing the array.

   The thresholding code does not work well on images
   with alpha channels (i.e. translucent parts).

   Usage:
     ToChars.bat p2_01.bmp
     ToChars.bat blueTele.png 20 20
*/


import java.io.*;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;



public class ToChars
{
  // dimensions of the Dream Cheeky LED Message Board
  private static final int SCR_WIDTH = 21;
  private static final int SCR_HEIGHT = 7;


  public static void main(String[] args)
  {
    if (!((args.length == 1) || (args.length == 3))) {
      System.out.println("Usage: ToChars <image-fnm> [<width> <height>]");
      return;
    }

    try {
      BufferedImage im = ImageIO.read(new File(args[0]));
      int imWidth = im.getWidth();
      int imHeight = im.getHeight();
      System.out.println("Orig. Dimensions (w, h): (" + 
                                    imWidth + ", " + imHeight + ")");

      // perhaps scale the image
      BufferedImage scaledIm;
      if (args.length == 1) {
        if ((imWidth == SCR_WIDTH) && (imHeight == SCR_HEIGHT)) {
          System.out.println("No scaling being applied");
          scaledIm = im;
        }
        else  // use Dream Cheeky dimensions
          scaledIm = scaleImage(im, SCR_WIDTH, SCR_HEIGHT);
      }
      else
        scaledIm = scaleImage(im, parseInt(args[1]), parseInt(args[2]));

      // use thresholding to convert the image to black and white
      BufferedImage bwIm = thresholdImage(scaledIm, 254);
        /* The threshold is very close to full-white (255) so that
           even light grays will be mapped to black.  */

      // convert the image to a char array and print it
      printArray(im2Chars(bwIm));
    }
    catch (Exception e) 
    {  System.out.println("Unable to process " + args[0]); }
  }  // end of main()


  private static int parseInt(String s)
  {
    int val = 0;
    try {
      val = Integer.parseInt(s);
    }
    catch (NumberFormatException e){ 
      System.out.println("Could not extract int from " + s); 
    }
    return val;
  }  // end of parseInt()



  private static BufferedImage scaleImage(BufferedImage im, int w, int h)
  // return a new image scaled to (w * h)
  { 
    int imWidth = im.getWidth();
    int imHeight = im.getHeight();

    // apply same scaling factor to both dimensions;
    // decide on factor by examining original dimensions
    double factor = 1.0d;
    if (imWidth > imHeight){
      factor = ((double)imHeight/(double)imWidth);
      h = (int)(w * factor);                
    }
    else {
      factor = ((double)imWidth/(double)imHeight);
      w = (int)(h * factor);
    }   

    BufferedImage imScaled = new BufferedImage(w, h, im.getType());
    Graphics2D g2 = imScaled.createGraphics();
    g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, 
                        RenderingHints.VALUE_INTERPOLATION_BILINEAR);
    g2.drawImage(im, 0, 0, w, h, null);
    g2.dispose();
    return imScaled;
  }  // end of scaleImage()



  private static BufferedImage thresholdImage(BufferedImage im, int threshold) 
  /* Create a grayscale copy of the image, and then examine each pixel.
     If a pixel is below a threshold, set to black, otherwise white.  
  */
  {
    int imWidth = im.getWidth();
    int imHeight = im.getHeight();

    // copy the image as a grayscale
    BufferedImage imThresh = new BufferedImage(imWidth, imHeight, 
                                           BufferedImage.TYPE_BYTE_GRAY);
    imThresh.getGraphics().drawImage(im, 0, 0, null);

    // examine pixels row-by-row
    int[] pixels = new int[imWidth];
    WritableRaster raster = imThresh.getRaster();
    for (int y = 0; y < imHeight; y++) {
      raster.getPixels(0, y, imWidth, 1, pixels);  // get a row
      for (int i = 0; i < pixels.length; i++) {
        if (pixels[i] < threshold) 
          pixels[i] = 0;    // set to black
        else 
          pixels[i] = 255;  // white
      }
      raster.setPixels(0, y, imWidth, 1, pixels);
    }
    return imThresh;
  }  // end of thresholdImage()




  private static char[][] im2Chars(BufferedImage im)
  // Convert an image into a char array
  {
    int imWidth = im.getWidth();
    int imHeight = im.getHeight();
    char[][] chArr = new char[imHeight][imWidth];

    for (int r = 0; r < imHeight; r++) {
      for (int c = 0; c < imWidth; c++) {
        int pixel = im.getRGB(c,r);
      //  printPixel(pixel);
        if ( isBlack(pixel) )
          chArr[r][c] = '*';   // for black
        else 
          chArr[r][c] = ' ';   // for white
      }
    }
    return chArr;
  }  // end of im2Chars()



  private static void printPixel(int pixel) 
  // Print a pixel in ARGB format
  {
    int alpha = (pixel >> 24) & 0xff;
    int red = (pixel >> 16) & 0xff;
    int green = (pixel >> 8) & 0xff;
    int blue = (pixel) & 0xff;
    System.out.print("(" + alpha + ", " + red + ", " + green + ", " + blue + ") ");
  }


  private static boolean isBlack(int pixel)
  {
    int red = (pixel >> 16) & 0xff;
    int green = (pixel >> 8) & 0xff;
    int blue = (pixel) & 0xff;
    return ((red == 0) && (green == 0) && (blue == 0));
  }


  private static void printArray(char[][] charArr)
  {
    if (charArr == null) {
      System.out.println("Char array is null");
      return;
    }

    System.out.println("Dimensions (w, h): (" + charArr[0].length +
                                         ", " + charArr.length + ")");
    System.out.println("-----------------------");
    for (int r = 0; r < charArr.length; r++) {
      for (int c = 0; c < charArr[0].length; c++) {
        System.out.print( charArr[r][c]);
      }
      System.out.println();
    }
    System.out.println("-----------------------");
  }  // end of printArray()



}  // end of ToChars class
