
// ToHex.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th


/* Convert char arrays into encoded hexadecimals suitable
   for reading in as a font using 
   SpriteStore.loadFont(<fnm>, false);

   false is used since the array is encoded as one hexadecimal
   per row. So 7 hexadecimals are output, such as:

     /   0x00  0x08  0x04  0x02  0x01  0x00  0x00
     .   0x00  0x00  0x00  0x00  0x06  0x06  0x00
     A   0x06  0x09  0x0F  0x09  0x09  0x00  0x00

   The char array is assumed to be 5 x 7 (columns x rows).

   Usage:
      ToHex.bat chars1.txt
*/

import java.io.*;
import java.awt.*;
import java.util.*;



public class ToHex
{
  // size of a char array for a letter pic
  private static final int CHAR_WIDTH = 5;
  private static final int CHAR_HEIGHT = 7;    
               // same as Dream Cheeky Message Board height



  public static void main(String[] args)
  {
    if (args.length != 1)
      System.out.println("Usage: ToChars <chars-fnm>");
    else
      loadChars(args[0]);
  } 


  private static void loadChars(String fnm)
  /* The format of the input lines are:
          $$ <letter>   
          // followed by a char array starting on the next line
          -- each char array is assumed to be CHAR_WIDTH chars wide 
             and CHAR_HEIGHT rows deep
          -- the array lines consist of spaces and *'s

     The file can also contain and empty lines and comment lines.

     There may be many letter defs in the file.
  */
  { 
    char letter = ' ';
    char[][] chArr = new char[CHAR_HEIGHT][CHAR_WIDTH];
    int rowNo = 0;
    boolean isFillingArr = false;
    int count = 0;

    System.out.println("Loading chars from " + fnm);
    try {
      BufferedReader br = new BufferedReader( new FileReader(fnm));
      String line;
      while((line = br.readLine()) != null) {
        if ( (line.length() == 0) &&    // empty line outside of char array
             (!isFillingArr))
          continue;
        if (line.startsWith("//"))   // comment
          continue;

        if (line.startsWith("$$")) {    // a single letter char name
          if (isFillingArr)    // but still filling previous array?
            System.out.println("Missing data for '" + 
                                     letter + "'; not converted");
          // start processing next char
          letter = line.charAt(3);   // should error check this
          ArrUtils.clear(chArr);
          isFillingArr = true;
          rowNo = 0;
        }
        else if (isFillingArr) {  // filling a char array
          fillRow(chArr, rowNo, line);
          rowNo++;
          if (rowNo == CHAR_HEIGHT) {  // finished reading data
            convert2Hex(letter, chArr);
            isFillingArr = false;
            count++;
          }
        }
      }

      if (isFillingArr)
        System.out.println("Missing data for '" + 
                               letter + "'; not converted");
      br.close();
    } 
    catch (IOException e) 
    { System.out.println("Error reading characters info file: " + fnm);   }

    System.out.println("No. of characters converted: " + count);
  }  // end of loadChars()




  private static void fillRow(char[][] chArr, int rowNo, String line)
  // store the line in the specified row of the char array
  {
    int width = chArr[0].length;

    if (line.length() < width) {
      // System.out.println("Line is short, adding spaces to \"" + line + "\"");
      width = line.length();
    }
    else if (line.length() > width)
      System.out.println("Line too long, only processing first " + 
                                  width + " chars in \"" + line + "\"");

    for (int i=0; i < width; i++) {
      char ch = line.charAt(i);
      if ((ch == '*') || (ch == ' '))
         chArr[rowNo][i] = ch;
      else {
        System.out.println("Unknown char: '" + ch +"'; using '*'");
        chArr[rowNo][i] = '*';
      }
    }
  }  // end of fillRow();



  private static void convert2Hex(char letter, char[][] chArr)
  {
    if (chArr == null) {
      System.out.println("Array is null");
      return;
    }

    int[] rowVals = new int[CHAR_HEIGHT];
    for (int r = 0; r < CHAR_HEIGHT; r++) {
      int val = 0;
      int power = 1;
      for (int c = 0; c < CHAR_WIDTH; c++) {
        if (chArr[r][c] == '*')
          val += power;
        power *= 2;
      }
      rowVals[r] = val;
    }

    System.out.print(letter + "   ");
    for (int i = 0; i < CHAR_HEIGHT; i++)
       System.out.printf("0x%02X  ", rowVals[i]);
    System.out.println();
  }  // end of convert2Hex()



}  // end of ToHex class
