
// Utils.java
// Andrew Davison, March 2017, ad@fivedots.coe.psu.ac.th


/* The main sections of this class are:

    * sprite effects
    * multi-picture sprite effects
    * animate a sprite by changing its position
    * generate points making up a shape
        -- line, rectangle, circle
    * sound-related methods
    * other utilities
*/

import java.awt.*;
import java.util.*;

import java.net.*;
import javax.sound.sampled.*;
import java.util.zip.*;



public class Utils
{
  private static final String SND_DIR = "sounds/";
       // where the WAV files are stored relative to this directory



  // ----------- sprite effects -------------


  public static void flash(Screen scr, Sprite s, int numTimes)
  // make this sprite appear/disappear a number of times
  {
    for (int i =0; i < numTimes; i++) {
      scr.add(s);
      scr.draw();
      delay(500);

      scr.clear(s);  // only this sprite is wiped
      scr.draw();
      delay(200);
    }
  }  // end of flash()



  public static void flashInvert(Screen scr, Sprite s, int num)
  // make this sprite flash by inverting its picture a few times
  {
    int numTimes = num*2;
    // s.backup();
    for(int i=0; i < numTimes; i++) {
      scr.clear(s);
      s.transform("invert");
      scr.add(s);
      scr.draw();
      Utils.delay(200);        
    }
    // s.restore();
  }  // end of flashInvert()



  public static void showAll(Screen scr, Sprite[] sprites, int numTimes)
  /* Display each sprite in the array in turn, repeating the specified
     number of times. */
  {
    for (int i=0; i < numTimes; i++) {
      for (int j=0; j < sprites.length; j++) {
        scr.add(sprites[j]);
        scr.draw();
        Utils.delay(200);
        scr.clear(sprites[j]);
      }
    }
  }  // end of showAll()



 // ----- effects using a multi-picture sprite -------


  public static void loop(Screen scr, Sprite s, int numTimes)
  // cycle through the sprite's pictures a specified number of times
  {
    int numPics = s.numPics();
    for(int i=0; i < (numTimes * numPics); i++) {
      scr.clear(s);
      s.toPic(i%numPics);
      scr.add(s);
      scr.draw();
      delay(200);
    }
  }  // end of loop()



  public static void showSeq(Screen scr, Sprite s)
  {  loop(scr, s, 1);  }



  public static Sprite buildUp(Sprite s)
  /* Create a multi-picture sprite by generating multiple
     pictures from the character array in the supplied sprite.
     Each picture is a slice of the array by increasing the
     number of rows in the slice counting from the bottom of
     the array.
  */
  { ArrayList<char[][]> pics = ArrUtils.buildUp(s.getCharArray());  
    return new Sprite(pics);
  }



  public static Sprite buildRight(Sprite s)
  /* Create a multi-picture sprite by generating multiple
     pictures from the character array in the supplied sprite.
     Each picture is a slice of the array by increasing the
     number of columns in the slice counting from the left of
     the array.
  */
  { ArrayList<char[][]> pics = ArrUtils.buildRight(s.getCharArray());  
    return new Sprite(pics);
  }



  // -------- animate a sprite by changing its position -----------


  public static void scrollLeft(Screen scr, Sprite s)
  /* Move the sprite right to left, starting off the right
     edge and finishing off the left edge.
  */
  { int x0 = scr.getWidth();
    int x1 = -s.getWidth();
    moveBetween(scr, s, x0, 0, x1, 0);
  }  // end of scrollLeft()


  public static void scrollUp(Screen scr, Sprite s)
  /* Move the sprite upwards, starting off the bottom 
     edge and finishing off the top edge. It is positioned
     in the center of the x-axis.
  */
  { int xMid = (int) Math.round((scr.getWidth() - s.getWidth())/2);
    int y0 = scr.getHeight();
    int y1 = -s.getHeight();
    moveBetween(scr, s, xMid, y0, xMid, y1);
  }  // end of scrollUp()



  public static void moveBetween(Screen scr, Sprite s,
                          int x0, int y0, int x1, int y1)
  /* Use calcLine() to generate a list of points in a line
     between (x0,y0) and (x1,y1). 

     Use the points to change the position of the sprite inside
     a loop that keeps redrawing it at the new position.
  */
  {
    ArrayList<Point> path = calcLine(x0, y0, x1, y1);

    if (path.size() == 0) 
      System.out.println("No path generated");
    else {
      // move along path
      for (int i=0; i < path.size(); i++) {
        scr.clear(s);   // wipe the sprite only
        s.setPos(path.get(i));
        scr.add(s);
        scr.draw();
        delay(200);
      }
    }
  }  // end of moveBetween()



  // ---------- generate points making up a shape -----------------



  public static ArrayList<Point> calcLine(int x0, int y0, int x1, int y1)
  /* return the points making up a line connecting the two points 
     (x0,y0) and (x1,y1). 
     Uses the Bresenham line algorithm.
  */
  {
    ArrayList<Point> pts = new ArrayList<Point>();

    int dx = Math.abs(x1 - x0);
    int dy = Math.abs(y1 - y0);
    int sx = (x0 < x1) ? 1 : -1;
    int sy = (y0 < y1) ? 1 : -1;

    int err = dx - dy;

    pts.add( new Point(x0, y0));

    while (!((x0 == x1) && (y0 == y1))) {
      int e2 = 2 * err;
      if (e2 > -dy) {
        err = err - dy;
        x0 = x0 + sx;
      }

      if (e2 < dx) {
        err = err + dx;
        y0 = y0 + sy;
      }
      pts.add( new Point(x0,y0));
    }
    return pts;
  }  // end of calcLine()



  public static ArrayList<Point> calcRect(int x, int y, int w, int h)
  /*  Create a list of rectangle points, by generating the points
      for the 4 edges.  */
  {
    ArrayList<Point> pts = new ArrayList<Point>();

    ArrayList<Point> ptsLine = calcLine(x, y, x+w, y);  // across top
    pts.addAll(ptsLine);

    ptsLine = calcLine(x+w, y, x+w, y+h);  // down right
    pts.addAll(ptsLine);

    ptsLine = calcLine(x+w, y+h, x, y+h);  // across bottom
    pts.addAll(ptsLine);

    ptsLine = calcLine(x, y+h, x, y);  // up left
    pts.addAll(ptsLine);

    return pts;
  }  // end of calcRect()




  public static ArrayList<Point> calcCircle(int x, int y, int radius)
  // Generate points running around the circle, rounded to integers
  {
    int numPoints = (2*radius + 1) * 2;    // two diameters for perimeter
    ArrayList<Point> pts = new ArrayList<Point>();

    double slice = 2 * Math.PI / numPoints;
    for (int i = 0; i < numPoints; i++) {
      double angle = slice * i;
      int xc = (int)Math.round(x + radius*Math.cos(angle));
      int yc = (int)Math.round(y + radius*Math.sin(angle));
      Point p = new Point(xc, yc);
      // System.out.println(p);
      pts.add(p);
    }
    return pts;
  }  // end of calcCircle()




  // ------------ sound methods -------------------


  public static ArrayList<String> getSounds()
  // return a list of the ".wav" filenames in the directory
  {
    ArrayList<String> fnms = new ArrayList<String>();

    try {
      URL jar = Utils.class.getProtectionDomain().
                                     getCodeSource().getLocation();
      ZipInputStream zip = new ZipInputStream(jar.openStream());
      ZipEntry ze = null;
      while((ze = zip.getNextEntry()) != null) {
        String fnm = ze.getName();
        // System.out.println(fnm);
        if(fnm.startsWith(SND_DIR) && fnm.endsWith(".wav")) {
          int idx = fnm.lastIndexOf("/") + 1;
          fnms.add(fnm.substring(idx));
        }
      }
    }
    catch(Exception e)
    {  System.out.println("Could not access sounds: " + e);  }
    return fnms;
  }  // end of getSounds()



  public static void play(final String nm) 
  {  play(nm, 1);  }


  public static void play(final String nm, final int numTimes) 
  // play nm sound without waiting for it to finish;
  // play it numTimes
  {        
    Thread soundThread = new Thread() {
      public void run() 
      { playWait(nm, numTimes);  }
    };
    soundThread.setDaemon(true);
    soundThread.start();
  } // end of play()


  public static void playWait(String nm, int numTimes) 
  { for (int i=0; i < numTimes; i++)
      playWait(nm);
  } 



  public static void playWait(String nm) 
  // play nm sound and wait for it to finish;
  {
    String fnm = nm;
    if (nm.indexOf('.') == -1)   // no extension
      fnm= fnm + ".wav";

    try {
      URL url  = Utils.class.getResource("/" + SND_DIR + fnm);
      // System.out.println(url);
      AudioInputStream ais = AudioSystem.getAudioInputStream(url);
      Clip sndClip = AudioSystem.getClip();  
      sndClip.open(ais);
      sndClip.start();
      //System.out.println("Playing " + fnm);

      while (!sndClip.isRunning())
        Thread.sleep(10);
      while (sndClip.isRunning())
        Thread.sleep(10);

      sndClip.close();
    }
    catch(Exception e){
      System.out.println("Unable to play " + fnm);
    }
  }  // end of playWait()



  // ------------ other utilities -------------------


  public static void delay(int ms)
  // delay execution for the specified number of milliseconds
  { try {
      Thread.sleep(ms);
    }
    catch(InterruptedException e){}
  }


  public static void waitEnter()
  // delay execution until the use presses <enter>
  {
    System.out.println("Press Enter to continue...");
    try {
      System.in.read();
    }
    catch(java.io.IOException e){}
  }  // end of waitEnter()


}  // end of Utils class